/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2011 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.

 compile:
g++ -I/home/spanderen/workspace/QuantLib -I/usr/local/include/boost-1_38 -O3 -o vpp vpp2.cpp -L/home/spanderen/workspace/QuantLib/ql/.libs -lQuantLib -lglpk 
*/

#include <ql/math/functional.hpp>
#include <ql/time/daycounters/actualactual.hpp>
#include <ql/math/randomnumbers/rngtraits.hpp>
#include <ql/math/statistics/generalstatistics.hpp>
#include <ql/experimental/processes/extouwithjumpsprocess.hpp>
#include <ql/experimental/processes/extendedornsteinuhlenbeckprocess.hpp>

#include <glpk.h>
#include <fstream>

using namespace QuantLib;

namespace {

    // model definition
    const Real beta         = 200;
    const Real eta          = 1.0/0.4;
    const Real lambda       = 4.0;
    const Real alpha        = 7.0;
    const Real volatility_x = 1.4;
    const Real kappa        = 4.45;
    const Real volatility_u = std::sqrt(1.3);
    const Real rho          = 0.7;
    const Rate irRate       = 0.00;

    // vpp definition
    const Size maturityInWeeks = 52;
    const Real efficiencies[] 
        = { 0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.60 }; // MWh power per MWh heat
    const Real pMin             = 8;    // MWh
    const Real pMax             = 40;   // MWh
    const Real startUpFuel      = 20;   // MWh
    const Real startUpFixedCost = 300;  // EUR
    const Real carbonPrice      = 3.0;  // EUR per MWh heat (from gas)
    const Size minOn            = 2.0;  // hour
    const Size minOff           = 2.0;  // hour

    glp_prob* createGLPKProblem(Size n, Size minOn, Size minOff) {

        glp_prob *lp;
        lp = glp_create_prob();
        
        char varName[32]; 

        glp_set_prob_name(lp, "vpp pricing");
        glp_set_obj_dir(lp, GLP_MAX);

        glp_add_cols(lp, 3*n);

        for (Size i=1; i <= n; ++i) {
            sprintf(varName,"running_%u",i);            
            glp_set_col_name(lp, i, varName);
            glp_set_col_kind(lp, i, GLP_BV);

            sprintf(varName,"start_%u",i);            
            glp_set_col_name(lp, n+i, varName);
            glp_set_col_bnds(lp, n+i, GLP_DB, 0.0, 1.0);

            sprintf(varName,"p_max_%u",i);
            glp_set_col_name(lp, 2*n+i, varName);
            glp_set_col_bnds(lp, 2*n+i, GLP_DB, 0.0, 1.0);
        }

        const Size nRows = 4*n;
        glp_add_rows(lp, nRows);

        std::vector<double> ar(1);
        std::vector<int> ia(1), ja(1);
        const Size reserve = (7 + minOff + minOn)*n+1;
        ar.reserve(reserve);
        ia.reserve(reserve);
        ja.reserve(reserve);

        for (Size i=1; i <= n; ++i) {
            // startUp: running(t) - running(t-1) - start(t) <= 0
            int row = i;
            sprintf(varName,"start_constraint_%u",i);
            glp_set_row_name(lp, row, varName);
            glp_set_row_bnds(lp, row, GLP_UP, 0.0, 0.0);

            ia.push_back(row); ja.push_back(i); ar.push_back(1.0);

            if (i > 1) {
                ia.push_back(row); ja.push_back(i-1); ar.push_back(-1.0);
            }
            ia.push_back(row); ja.push_back(i+n); ar.push_back(-1.0);

            //minOn: running(t) >= sum_{t-t_min+1}^t start(t)
            row = n+i;
            sprintf(varName,"minOn_constraint_%u",row);
            glp_set_row_name(lp, row, varName);
            glp_set_row_bnds(lp, row, GLP_LO, 0.0, 0.0);

            ia.push_back(row); ja.push_back(i); ar.push_back(1.0);
            for (int j=i-std::min(i, minOn)+1; j <=i; ++j) {
                ia.push_back(row); ja.push_back(j+n); ar.push_back(-1.0);
            }

            //minOff running(t) <= 1-sum_{t+1}^{t+t_min} start(t)
            row = i+2*n;
            sprintf(varName,"minOff_constraint_%u",row);
            glp_set_row_name(lp, row, varName);
            glp_set_row_bnds(lp, row, GLP_UP, 0.0, 1.0);

            ia.push_back(row); ja.push_back(i); ar.push_back(1.0);
            for (int j=i+1; j <= std::min(i+minOff, n); ++j) {
                ia.push_back(row); ja.push_back(j+n); ar.push_back(1.0);
            }

            // running(i) >= p_max(i)
            row = 3*n+i;
            sprintf(varName,"p_max_constraint_%u",i);
            glp_set_row_name(lp, row, varName);
            glp_set_row_bnds(lp, row, GLP_UP, 0.0, 0.0);

            ia.push_back(row); ja.push_back(i);     ar.push_back(-1.0);
            ia.push_back(row); ja.push_back(2*n+i); ar.push_back(1.0);
        }

        // see http://herbsutter.com/2008/04/07/cringe-not-vectors-are-\
        //       guaranteed-to-be-contiguous/
        glp_load_matrix(lp, ia.size()-1, &ia[0], &ja[0], &ar[0]);

        return lp;
    }

    void deleteGLPKProblem(glp_prob* lp) {
        glp_delete_prob(lp);
    }

    struct VPPResult {
        Real price;
        Size onTime;
        Size onTimePmax;
        Size numberOfStarts;
    };

    VPPResult pathPricer(glp_prob* lp, 
                         const std::vector<Real> eexPrice,
                         const std::vector<Real> ttfPrice,
                         Real efficiency, 
                         Real startUpFuel, Real startUpFixedCost, 
                         Real pMin, Real pMax, Real carbonPrice) {

        const Size n = eexPrice.size();

        for (Size i=0; i < n; ++i) {
            const Real exerciseValue 
                = eexPrice[i] - (ttfPrice[i]+carbonPrice)/efficiency;

            const Real pMinValue = pMin*exerciseValue;
            const Real pMaxValue = (pMax-pMin)*exerciseValue;
            const Real startUpCost 
                = startUpFixedCost + startUpFuel*(ttfPrice[i]+carbonPrice);

            glp_set_obj_coef(lp, i+1, pMinValue);
            glp_set_obj_coef(lp,   eexPrice.size()+i+1, -startUpCost);
            glp_set_obj_coef(lp, 2*eexPrice.size()+i+1, pMaxValue);
        }

        glp_iocp iocp;
        glp_init_iocp(&iocp);
        iocp.presolve=1;
        iocp.msg_lev = GLP_MSG_ERR;

        glp_intopt(lp, &iocp);
        QL_REQUIRE(glp_mip_status(lp) == GLP_OPT 
                   || glp_mip_status(lp) == GLP_FEAS, 
                   "infeasable problem");

        
        Size onTime = 0;
        Size onTimePmax = 0;
        Size numberOfStarts = 0;
        
        for (Size i=1; i <= n; ++i) {
            if (glp_mip_col_val(lp,     i) > 0) {
                ++onTime;
            }
            if (glp_mip_col_val(lp,   n+i) > 0) {
                ++numberOfStarts;
            }
            if (glp_mip_col_val(lp, 2*n+i) > 0) {
                ++onTimePmax;
            }
        }

        VPPResult result = { glp_mip_obj_val(lp), 
                             onTime, onTimePmax, numberOfStarts };
        
        return result;
    }
}




int main(int argc, char *argv[]) {
    const DayCounter dc = ActualActual();
    const Date settlementDate = Date(28, Sep, 2010);
    Settings::instance().evaluationDate() = settlementDate;

    const Size maturityInDays  = 7*maturityInWeeks;
    const Size maturityInHours = 24*maturityInDays;
    Date maturityDate = settlementDate + Period(maturityInWeeks, Weeks);

    std::vector<Time> exerciseTimes;
    exerciseTimes.reserve(maturityInHours);

    std::vector<Real> eexPrices, ttfPrices;
    std::vector<Real> eexShape, ttfShape;
    eexShape.reserve(maturityInHours);
    ttfShape.reserve(maturityInHours);
    eexPrices.reserve(maturityInHours);
    ttfPrices.reserve(maturityInHours);

    // read shape from file
    // format is Date, 0h price, 1h price, 2h price, ... 23h price
    // example:4-Oct-2010 36.73 32.09 27.32 23.22 25.71 35.60 47.32 ...
    std::ifstream eex("EEX_2010-2013.txt");
    QL_REQUIRE(eex.is_open(), "can't find price file EEX_2010-2013.txt");
    
    std::ifstream ttf("TTF_2010-2013.txt");
    QL_REQUIRE(ttf.is_open(), "can't find price file TTF_2010-2013.txt");

    Real p;
    for (Size i=0; i < maturityInDays; ++i) {
        std::string str;
        eex >> str;
        ttf >> str;
        
        Date date = settlementDate+Period(i, Days);
        for (Size j=0; j < 24; ++j) {
            Time t = dc.yearFraction(settlementDate, date);
            const Time dt = dc.yearFraction(settlementDate, 
                                            date+Period(1u, Days))-t;
            t += j*dt/24.;
            exerciseTimes.push_back(t);
            
            eex >> p;
            eexPrices.push_back(p);
            eexShape.push_back(std::log(p)-square<Real>()(volatility_x)
                 /(4*alpha)*(1-std::exp(-2*alpha*t))
                -lambda/beta*std::log((eta-std::exp(-beta*t))/(eta-1.0)));
            
            ttf >> p;
            ttfPrices.push_back(p);
            ttfShape.push_back(std::log(p)-square<Real>()(volatility_u)
                               /(4*kappa)*(1-std::exp(-2*kappa*t)));
        }
    }
    eex.close();
    ttf.close();
 
    glp_prob* lp = createGLPKProblem(maturityInHours, minOn, minOff);

    printf("Intrinsic Value:\n");
    for (Size i=0; i < sizeof(efficiencies)/sizeof(Real); ++i) {
        const VPPResult result 
            = pathPricer(lp, eexPrices, ttfPrices, 
                         efficiencies[i], 
                         startUpFuel,  startUpFixedCost, 
                         pMin, pMax, carbonPrice);

        printf("efficiency: %2.0f%%  price: %7.0f  OnTime: %4u  "
               "PmaxTime: %4u  #starts: %3u\n", 
               efficiencies[i]*100, result.price, 
               result.onTime, result.onTimePmax, result.numberOfStarts);
    }


    printf("\nExtrinsic Value (perfect foresight):\n");

    Array x0(2);
    x0[0] = 0.0; x0[1] = 0.0;

    TimeGrid grid(exerciseTimes.begin(), exerciseTimes.end(), 1);
    boost::shared_ptr<ExtendedOrnsteinUhlenbeckProcess> ouProcess(
        new ExtendedOrnsteinUhlenbeckProcess(alpha, volatility_x, x0[0],
                                             constant<Real, Real>(x0[0])));
    boost::shared_ptr<ExtOUWithJumpsProcess> lnPowerProcess(
           new ExtOUWithJumpsProcess(ouProcess, x0[1], beta, lambda, eta));
    
    Real u=0.0;
    boost::shared_ptr<ExtendedOrnsteinUhlenbeckProcess> lnGasProcess(
        new ExtendedOrnsteinUhlenbeckProcess(kappa, volatility_u, u,
                                             constant<Real, Real>(u)));
    
    const Time dt       = 1/(365*24.);
    const Time sqrtDt   = std::sqrt(dt);
    const Real dSqrtRho = std::sqrt(1-rho*rho);

    std::vector<Real> eexSim(maturityInHours), ttfSim(maturityInHours);

    const Size nTrails = 512;
    for (Size i=0; i < sizeof(efficiencies)/sizeof(Real); ++i) {
        PseudoRandom::rng_type rng 
            = PseudoRandom::rng_type(PseudoRandom::urng_type(11234567ul));

        GeneralStatistics npv, onTime, onTimePmax, numberOfStarts;

        for (Size j=0; j < nTrails; ++j) {
            for (Size k=0; k < maturityInHours; ++k) {
                const Time t = k/(365.*24);
                
                Array z(3);
                z[0] = rng.next().value;
                z[1] = rng.next().value;
                z[2] = rng.next().value;
                const Real z4 = rho*z[0] + dSqrtRho*rng.next().value;
                
                x0 = lnPowerProcess->evolve(t, x0, dt, z);
                u = lnGasProcess->evolve(t, u, dt, z4);
                
                ttfSim[k] = std::exp(u+ttfShape[k]);
                eexSim[k] = std::exp(x0[0]+x0[1]+eexShape[k]);
            }

            const VPPResult result 
                = pathPricer(lp, eexSim, ttfSim, 
                             efficiencies[i], 
                             startUpFuel,  startUpFixedCost, 
                             pMin, pMax, carbonPrice);

            npv.add(result.price);
            onTime.add(result.onTime);
            onTimePmax.add(result.onTimePmax);
            numberOfStarts.add(result.numberOfStarts);

            printf("pricing scenario %u\r", j);
            fflush(stdout);
        }

        printf("efficiency: %2.0f%%  price: %7.0f(%5.0f)  "
               "OnTime: %4.0f(%2.0f)  "
               "PmaxTime: %4.0f(%2.0f)  #starts: %3.0f(%2.1f)\n", 
               efficiencies[i]*100, 
               npv.mean(),            npv.errorEstimate(),
               onTime.mean(),         onTime.errorEstimate(),
               onTimePmax.mean(),     onTimePmax.errorEstimate(),
               numberOfStarts.mean(), numberOfStarts.errorEstimate());
    }



    deleteGLPKProblem(lp);

    return 0;
}
